/*
 * Austin---Astonishing Universal Search Tree Interface Novelty
 * Copyright (C) 2000 Kaz Kylheku <kaz@ashi.footprints.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * This source code may be translated into executable form and incorporated
 * into proprietary software; there is no requirement for such software to
 * contain a copyright notice related to this source.
 *
 * $Id: list.c,v 1.4 1999/11/27 02:58:21 kaz Exp $
 * $Name: austin_0_1 $
 */

#include <assert.h>
#include "udict.h"
#include "macros.h"

#define list_first_priv(L) ((L)->sentinel.next)
#define list_last_priv(L) ((L)->sentinel.prev)
#define list_null_priv(L) (&(L)->sentinel)

static void insert_before(udict_t *ud, udict_node_t *newnode, udict_node_t *succ)
{
    udict_node_t *pred = succ->prev;

    newnode->prev = pred;
    newnode->next = succ;

    pred->next = newnode;
    succ->prev = newnode;

    assert (ud->nodecount < ud->maxcount);
    ud->nodecount++;
}

static void list_init(udict_t *ud)
{
    ud->sentinel.next = &ud->sentinel;
    ud->sentinel.prev = &ud->sentinel;
}

static void list_insert(udict_t *ud, udict_node_t *node, const void *key)
{
    udict_node_t *succ;

    node->key = key;

    for (succ = list_first_priv(ud); succ != list_null_priv(ud); succ = succ->next) {
	if (ud->compare(succ->key, key) > 0)
	    break;
    }

    insert_before(ud, node, succ);
}

static void list_delete(udict_t *ud, udict_node_t *node)
{
    udict_node_t *pred = node->prev;
    udict_node_t *succ = node->next;

    pred->next = succ;
    succ->prev = pred;
    assert (ud->nodecount > 0);
    ud->nodecount--;
}

static udict_node_t *list_lookup(udict_t *ud, const void *key)
{
    udict_node_t *node;

    for (node = list_first_priv(ud); node != list_null_priv(ud); node = node->next) {
	if (ud->compare(node->key, key) == 0)
	    return node;
    }

    return 0;
}

static udict_node_t *list_lower_bound(udict_t *ud, const void *key)
{
    udict_node_t *node;

    for (node = list_first_priv(ud); node != list_null_priv(ud); node = node->next) {
	if (ud->compare(node->key, key) >= 0)
	    return node;
    }

    return 0;
}

static udict_node_t *list_upper_bound(udict_t *ud, const void *key)
{
    udict_node_t *node;

    for (node = list_last_priv(ud); node != list_null_priv(ud); node = node->prev) {
	if (ud->compare(node->key, key) <= 0)
	    return node;
    }

    return 0;
}

static udict_node_t *list_first(udict_t *ud)
{
    return (list_first_priv(ud) == list_null_priv(ud)) ? 0 : list_first_priv(ud);
}

static udict_node_t *list_last(udict_t *ud)
{
    return (list_last_priv(ud) == list_null_priv(ud)) ? 0 : list_last_priv(ud);
}

static udict_node_t *list_next(udict_t *ud, udict_node_t *node)
{
    return (node->next == list_null_priv(ud)) ? 0 : node->next;
}

static udict_node_t *list_prev(udict_t *ud, udict_node_t *node)
{
    return (node->prev == list_null_priv(ud)) ? 0 : node->prev;
}

static void list_convert_to_list(udict_t *ud)
{
    /* already a list */
}

static void list_convert_from_list(udict_t *ud)
{
    /* already a list */
}

udict_operations_t udict_list_ops = {
    list_init,
    list_insert,
    list_delete,
    list_lookup,
    list_lower_bound,
    list_upper_bound,
    list_first,
    list_last,
    list_next,
    list_prev,
    list_convert_to_list,
    list_convert_from_list,
    udict_list
};
