/*
 * Austin---Astonishing Universal Search Tree Interface Novelty
 * Copyright (C) 2000 Kaz Kylheku <kaz@ashi.footprints.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * This source code may be translated into executable form and incorporated
 * into proprietary software; there is no requirement for such software to
 * contain a copyright notice related to this source.
 *
 * $Id: redblack.c,v 1.4 1999/12/09 05:38:52 kaz Exp $
 * $Name: austin_0_1 $
 */

#include <assert.h>
#include "udict.h"
#include "tree.h"
#include "macros.h"

#define color udict_algo_specific.udict_rb_color

static void redblack_init(udict_t *ud)
{
    udict_tree_init(ud);
    ud->sentinel.color = udict_black;
}

static void redblack_insert(udict_t *ud, udict_node_t *node, const void *key)
{
    udict_node_t *parent;

    /* basic BST insert */

    udict_tree_insert(ud, node, key);

    /* red-black fixup */

    node->color = udict_red;
    parent = node->parent;

    while (parent->color == udict_red) {
	udict_node_t *grandpa = parent->parent;

	if (parent == grandpa->left) {
	    udict_node_t *uncle = grandpa->right;
	    if (uncle->color == udict_red) {	/* red parent, red uncle */
		parent->color = udict_black;
		uncle->color = udict_black;
		grandpa->color = udict_red;
		node = grandpa;
		parent = grandpa->parent;
	    } else {				/* red parent, black uncle */
	    	if (node == parent->right) {
		    udict_tree_rotate_left(node, parent);
		    parent = node;
		    assert (grandpa == parent->parent);
		    /* rotation between parent and child preserves grandpa */
		}
		parent->color = udict_black;
		grandpa->color = udict_red;
		udict_tree_rotate_right(parent, grandpa);
		break;
	    }
	} else { 	/* symmetric cases: parent == parent->parent->right */
	    udict_node_t *uncle = grandpa->left;
	    if (uncle->color == udict_red) {
		parent->color = udict_black;
		uncle->color = udict_black;
		grandpa->color = udict_red;
		node = grandpa;
		parent = grandpa->parent;
	    } else {
	    	if (node == parent->left) {
		    udict_tree_rotate_right(node, parent);
		    parent = node;
		    assert (grandpa == parent->parent);
		}
		parent->color = udict_black;
		grandpa->color = udict_red;
		udict_tree_rotate_left(parent, grandpa);
		break;
	    }
	}
    }

    tree_root_priv(ud)->color = udict_black;
}

static void redblack_delete(udict_t *ud, udict_node_t *node)
{
    udict_node_t *swap, *child;
    udict_rb_color_t savecolor;

    /* basic BST delete */

    udict_tree_delete(ud, node, &swap, &child);
    
    /* In case node was replaced by successor, transfer color
       to successor */

    savecolor = node->color;
    node->color = swap->color;
    swap->color = savecolor;

    /* red-black fixup */

    if (node->color == udict_black) {
	udict_node_t *parent, *sister;

	tree_root_priv(ud)->color = udict_red;

	while (child->color == udict_black) {
	    parent = child->parent;
	    if (child == parent->left) {
		sister = parent->right;
		assert (sister != tree_null_priv(ud));
		if (sister->color == udict_red) {
		    sister->color = udict_black;
		    parent->color = udict_red;
		    udict_tree_rotate_left(sister, parent);
		    sister = parent->right;
		    assert (sister != tree_null_priv(ud));
		}
		if (sister->left->color == udict_black
			&& sister->right->color == udict_black) {
		    sister->color = udict_red;
		    child = parent;
		} else {
		    if (sister->right->color == udict_black) {
			assert (sister->left->color == udict_red);
			sister->left->color = udict_black;
			sister->color = udict_red;
			udict_tree_rotate_right(sister->left, sister);
			sister = parent->right;
			assert (sister != tree_null_priv(ud));
		    }
		    sister->color = parent->color;
		    sister->right->color = udict_black;
		    parent->color = udict_black;
		    udict_tree_rotate_left(sister, parent);
		    break;
		}
	    } else {	/* symmetric case: child == child->parent->right */
		assert (child == parent->right);
		sister = parent->left;
		assert (sister != tree_null_priv(ud));
		if (sister->color == udict_red) {
		    sister->color = udict_black;
		    parent->color = udict_red;
		    udict_tree_rotate_right(sister, parent);
		    sister = parent->left;
		    assert (sister != tree_null_priv(ud));
		}
		if (sister->right->color == udict_black
			&& sister->left->color == udict_black) {
		    sister->color = udict_red;
		    child = parent;
		} else {
		    if (sister->left->color == udict_black) {
			assert (sister->right->color == udict_red);
			sister->right->color = udict_black;
			sister->color = udict_red;
			udict_tree_rotate_left(sister->right, sister);
			sister = parent->left;
			assert (sister != tree_null_priv(ud));
		    }
		    sister->color = parent->color;
		    sister->left->color = udict_black;
		    parent->color = udict_black;
		    udict_tree_rotate_right(sister, parent);
		    break;
		}
	    }
	}

	child->color = udict_black;
	tree_root_priv(ud)->color = udict_black;
    }
}

/*
 * This is a cut and paste clone of of tree_convert_from_list(), modified
 * to colorize the nodes as it adds them to the tree. Nodes on an even
 * level are colored red, and nodes on an odd level are colored black.
 * The root node is unconditionally colored black.
 */

static void redblack_convert_from_list(udict_t *ud)
{
    udict_node_t *tree[TREE_DEPTH_MAX] = { 0 };
    udict_node_t *curr, *nil = tree_null_priv(ud), *next;
    udict_node_t *complete = 0;
    udict_count_t fullcount = UDICT_COUNT_T_MAX, nodecount = ud->nodecount;
    udict_count_t botrowcount;
    unsigned baselevel = 0, level = 0, i;

    assert (udict_red == 0 && udict_black == 1);

    while (fullcount >= nodecount && fullcount)
	fullcount >>= 1;

    botrowcount = nodecount - fullcount;

    for (curr = nil->next; curr != nil; curr = next) {
	next = curr->next;

	if (complete == 0 && botrowcount-- == 0) {
	    assert (baselevel == 0);
	    assert (level == 0);
	    baselevel = level = 1;
	    complete = tree[0];

	    if (complete != 0) {
		tree[0] = 0;
		complete->right = nil;
		while (tree[level] != 0) {
		    tree[level]->right = complete;
		    complete->parent = tree[level];
		    complete = tree[level];
		    tree[level++] = 0;
		}
	    }
	}

	if (complete == 0) {
	    curr->left = nil;
	    curr->right = nil;
	    curr->color = level % 2;
	    complete = curr;

	    assert (level == baselevel);
	    while (tree[level] != 0) {
		tree[level]->right = complete;
		complete->parent = tree[level];
		complete = tree[level];
		tree[level++] = 0;
	    }
	} else {
	    curr->left = complete;
	    curr->color = (level + 1) % 2;
	    complete->parent = curr;
	    tree[level] = curr;
	    complete = 0;
	    level = baselevel;
	}
    }

    if (complete == 0)
	complete = nil;

    for (i = 0; i < TREE_DEPTH_MAX; i++) {
	if (tree[i] != 0) {
	    tree[i]->right = complete;
	    complete->parent = tree[i];
	    complete = tree[i];
	}
    }

    nil->right = nil;
    nil->left = complete;
    nil->color = udict_black;
    complete->parent = nil;
    complete->color = udict_black;
}

udict_operations_t udict_redblack_ops = {
    redblack_init,
    redblack_insert,
    redblack_delete,
    udict_tree_lookup,
    udict_tree_lower_bound,
    udict_tree_upper_bound,
    udict_tree_first,
    udict_tree_last,
    udict_tree_next,
    udict_tree_prev,
    udict_tree_convert_to_list,
    redblack_convert_from_list,
    udict_bst
};
