/*
 * Austin---Astonishing Universal Search Tree Interface Novelty
 * Copyright (C) 2000 Kaz Kylheku <kaz@ashi.footprints.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * This source code may be translated into executable form and incorporated
 * into proprietary software; there is no requirement for such software to
 * contain a copyright notice related to this source.
 *
 * $Id: splay.c,v 1.3 1999/12/09 05:38:52 kaz Exp $
 * $Name: austin_0_1 $
 */

#include <assert.h>
#include "udict.h"
#include "tree.h"
#include "macros.h"

/*
 * The child node is a left child of its parent, which is a left child of the
 * grandparent. After this rotation, the child becomes the grandparent,
 * the parent becomes its right child and the original grandparent is
 * the right child of that. 
 */

static void right_zig_zig(udict_node_t *child, udict_node_t *parent, udict_node_t *grandpa)
{
    /*
     * For now we take the easy way out and do this using two rotations.
     * This can be optimized.
     */

    udict_tree_rotate_right(parent, grandpa);
    udict_tree_rotate_right(child, parent);
}

static void left_zig_zig(udict_node_t *child, udict_node_t *parent, udict_node_t *grandpa)
{
    udict_tree_rotate_left(parent, grandpa);
    udict_tree_rotate_left(child, parent);
}

/*
 * The child node is a left child of its parent, which is a right child of the
 * grandparent. After the adjustment, the child becomes a parent to the
 * grandparent and parent.
 */

static void right_zig_zag(udict_node_t *child, udict_node_t *parent, udict_node_t *grandpa)
{
    udict_tree_rotate_right(child, parent);
    udict_tree_rotate_left(child, grandpa);
}

static void left_zig_zag(udict_node_t *child, udict_node_t *parent, udict_node_t *grandpa)
{
    udict_tree_rotate_left(child, parent);
    udict_tree_rotate_right(child, grandpa);
}

static void splay_node(udict_t *ud, udict_node_t *node)
{
    udict_node_t *root = tree_root_priv(ud), *parent = node->parent;
    
    if (parent->left == node) {
	if (parent == root) {
	    udict_tree_rotate_right(node, parent);
	} else {
	    udict_node_t *grandpa = parent->parent;

	    if (grandpa->left == parent) {
		right_zig_zig(node, parent, grandpa);
	    } else {
		assert (grandpa->right == parent);
		right_zig_zag(node, parent, grandpa);
	    }
	}
    } else {
	assert (parent->right == node);

	if (parent == root) {
	    udict_tree_rotate_left(node, parent);
	} else {
	    udict_node_t *grandpa = parent->parent;

	    if (grandpa->right == parent) {
		left_zig_zig(node, parent, grandpa);
	    } else {
		assert (grandpa->left == parent);
		left_zig_zag(node, parent, grandpa);
	    }
	}

    }
}

static void splay_insert(udict_t *ud, udict_node_t *node, const void *key)
{
    udict_tree_insert(ud, node, key);

    while (node != tree_root_priv(ud))
	splay_node(ud, node);
}

static void splay_delete(udict_t *ud, udict_node_t *node)
{
    udict_node_t *dummy;
    udict_tree_delete(ud, node, &dummy, &dummy);
}

static udict_node_t *splay_lookup(udict_t *ud, const void *key)
{
    udict_node_t *node = udict_tree_lookup(ud, key);

    if (node)
	while (node != tree_root_priv(ud))
	    splay_node(ud, node);

    return node;
}

udict_operations_t udict_splay_ops = {
    udict_tree_init,
    splay_insert,
    splay_delete,
    splay_lookup,
    udict_tree_lower_bound,
    udict_tree_upper_bound,
    udict_tree_first,
    udict_tree_last,
    udict_tree_next,
    udict_tree_prev,
    udict_tree_convert_to_list,
    udict_tree_convert_from_list,
    udict_bst
};
