/*
 * Austin---Astonishing Universal Search Tree Interface Novelty
 * Copyright (C) 2000 Kaz Kylheku <kaz@ashi.footprints.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * This source code may be translated into executable form and incorporated
 * into proprietary software; there is no requirement for such software to
 * contain a copyright notice related to this source.
 *
 * $Id: tree.c,v 1.8 1999/12/09 05:38:52 kaz Exp $
 * $Name: austin_0_1 $
 */

#include <assert.h>
#include "udict.h"
#include "tree.h"
#include "macros.h"

void udict_tree_init(udict_t *ud)
{
    ud->sentinel.left = &ud->sentinel;
    ud->sentinel.right = &ud->sentinel;
    ud->sentinel.parent = &ud->sentinel;
    ud->sentinel.algo_specific.udict_dummy = 0;
    ud->sentinel.data = 0;
    ud->sentinel.key = 0;
}

void udict_tree_insert(udict_t *ud, udict_node_t *node, const void *key)
{
    udict_node_t *where = tree_root_priv(ud), *nil = tree_null_priv(ud);
    udict_node_t *parent = nil;
    int result = -1;

    node->key = key;

    while (where != nil) {
	parent = where;
	result = ud->compare(key, where->key);
	/* if this assert goes off, you're trying to insert duplicates
	   without calling udict_allow_dupes() first! */
	assert (ud->dupes_allowed || result != 0);
	if (result < 0)
	    where = where->left;
	else
	    where = where->right;
    }

    assert (where == nil);

    if (result < 0)
	parent->left = node;
    else
	parent->right = node;

    node->parent = parent;
    node->left = nil;
    node->right = nil;

    ud->nodecount++;
}

void udict_tree_delete(udict_t *ud, udict_node_t *node, udict_node_t **pswap, udict_node_t **pchild)
{
    udict_node_t *nil = tree_null_priv(ud), *child, *delparent = node->parent;
    udict_node_t *next = node, *nextparent;

    if (node->left != nil && node->right != nil) {
	next = udict_tree_next(ud, node);
	nextparent = next->parent;

	assert (next != nil);
	assert (next->parent != nil);
	assert (next->left == nil);

	/*
	 * First, splice out the successor from the tree completely, by
	 * moving up its right child into its place.
	 */

	child = next->right;
	child->parent = nextparent;

	if (nextparent->left == next) {
	    nextparent->left = child;
	} else {
	    assert (nextparent->right == next);
	    nextparent->right = child;
	}

	/*
	 * Now that the successor has been extricated from the tree, install it
	 * in place of the node that we want deleted.
	 */

	next->parent = delparent;
	next->left = node->left;
	next->right = node->right;
	next->left->parent = next;
	next->right->parent = next;

	if (delparent->left == node) {
	    delparent->left = next;
	} else {
	    assert (delparent->right == node);
	    delparent->right = next;
	}

    } else {
	assert (node != nil);
	assert (node->left == nil || node->right == nil);

	child = (node->left != nil) ? node->left : node->right;

	child->parent = delparent = node->parent;	    

	if (node == delparent->left) {
	    delparent->left = child;    
	} else {
	    assert (node == delparent->right);
	    delparent->right = child;
	}
    }

    node->parent = 0;
    node->right = 0;
    node->left = 0;

    ud->nodecount--;

    *pswap = next;
    *pchild = child;
}

static void tree_delete(udict_t *ud, udict_node_t *node)
{
    udict_node_t *dummy;
    udict_tree_delete(ud, node, &dummy, &dummy);
}

udict_node_t *udict_tree_lookup(udict_t *ud, const void *key)
{
    udict_node_t *root = tree_root_priv(ud);
    udict_node_t *nil = tree_null_priv(ud);
    udict_node_t *saved;
    int result;

    /* simple binary search adapted for trees that contain duplicate keys */

    while (root != nil) {
	result = ud->compare(key, root->key);
	if (result < 0)
	    root = root->left;
	else if (result > 0)
	    root = root->right;
	else {
	    if (!ud->dupes_allowed) {	/* no duplicates, return match		*/
		return root;
	    } else {		/* could be dupes, find leftmost one	*/
		do {
		    saved = root;
		    root = root->left;
		    while (root != nil && ud->compare(key, root->key))
			root = root->right;
		} while (root != nil);
		return saved;
	    }
	}
    }

    return 0;
}

udict_node_t *udict_tree_lower_bound(udict_t *ud, const void *key)
{
    udict_node_t *root = tree_root_priv(ud);
    udict_node_t *nil = tree_null_priv(ud);
    udict_node_t *tentative = 0;

    while (root != nil) {
	int result = ud->compare(key, root->key);

	if (result > 0) {
	    root = root->right;
	} else if (result < 0) {
	    tentative = root;
	    root = root->left;
	} else {
	    if (!ud->dupes_allowed) {
	    	return root;
	    } else {
		tentative = root;
		root = root->left;
	    }
	} 
    }
    
    return tentative;
}

udict_node_t *udict_tree_upper_bound(udict_t *ud, const void *key)
{
    udict_node_t *root = tree_root_priv(ud);
    udict_node_t *nil = tree_null_priv(ud);
    udict_node_t *tentative = 0;

    while (root != nil) {
	int result = ud->compare(key, root->key);

	if (result < 0) {
	    root = root->left;
	} else if (result > 0) {
	    tentative = root;
	    root = root->right;
	} else {
	    if (!ud->dupes_allowed) {
	    	return root;
	    } else {
		tentative = root;
		root = root->right;
	    }
	} 
    }
    
    return tentative;
}

udict_node_t *udict_tree_first(udict_t *ud)
{
    udict_node_t *nil = tree_null_priv(ud), *root = tree_root_priv(ud), *left;

    if (root != nil)
	while ((left = root->left) != nil)
	    root = left;

    return (root == nil) ? 0 : root;
}

udict_node_t *udict_tree_last(udict_t *ud)
{
    udict_node_t *nil = tree_null_priv(ud), *root = tree_root_priv(ud), *right;

    if (root != nil)
	while ((right = root->right) != nil)
	    root = right;

    return (root == nil) ? 0 : root;
}

udict_node_t *udict_tree_next(udict_t *ud, udict_node_t *curr)
{
    udict_node_t *nil = tree_null_priv(ud), *parent, *left;

    if (curr->right != nil) {
	curr = curr->right;
	while ((left = curr->left) != nil)
	    curr = left;
	return curr;
    }

    parent = curr->parent;

    while (parent != nil && curr == parent->right) {
	curr = parent;
	parent = curr->parent;
    }

    return (parent == nil) ? 0 : parent;
}

udict_node_t *udict_tree_prev(udict_t *ud, udict_node_t *curr)
{
    udict_node_t *nil = tree_null_priv(ud), *parent, *right;

    if (curr->left != nil) {
	curr = curr->left;
	while ((right = curr->right) != nil)
	    curr = right;
	return curr;
    }

    parent = curr->parent;

    while (parent != nil && curr == parent->left) {
	curr = parent;
	parent = curr->parent;
    }

    return (parent == nil) ? 0 : parent;
}

void udict_tree_convert_to_list(udict_t *ud)
{
    udict_node_t *node, tempsentinel, *nil = &tempsentinel;
    udict_node_t *tail, *next, *treenil = tree_null_priv(ud);

    if (ud->nodecount == 0)
	return;

    tempsentinel.next = nil;
    tempsentinel.prev = nil;

    /*
     * Two passes are required because udict_tree_next backtracks over
     * some predecessors when it hits a rightmost node of a left subtree. When
     * it does this, it needs to look at the right pointers as it goes up
     * through successive parents so that it knows when it's hit the parent of
     * the left subtree. Right pointers are overlaid with next pointers,
     * so on the first pass we only thread the list in the prev direction.
     * When it's built, we relink the next pointer.
     */

    for (tail = nil, node = udict_tree_first(ud); node != 0; tail = node, node = next) {
	next = udict_tree_next(ud, node);
	node->prev = tail;
    }

    nil->prev = tail;

    for (tail = nil, node = nil->prev; node != nil; tail = node, node = node->prev)
	node->next = tail;

    nil->next = tail;

    ud->sentinel.next = tempsentinel.next;
    ud->sentinel.prev = tempsentinel.prev;
    ud->sentinel.next->prev = treenil;
    ud->sentinel.prev->next = treenil;
}

/*
 * Implementation of slick algorithm for building nearly balanced binary search
 * tree from sorted list in O(N) time.  If the number of nodes is a Mersenne
 * number, a perfectly balanced tree is built. This algorithm constructs
 * ``half trees'' and keeps them in an array indexed by level. A half
 * tree is a node with no right child, and a fully balanced left child.
 * In this algorithm, a half tree's level is the height of the balanced child.
 * Whenever a full tree is constructed, the array is consulted to see if there
 * is a matching half tree of the same level; if so, the newly constructed
 * tree is installed as the right child of the half tree's root node. 
 * The result is a full tree, which is checked again for the possibility of
 * connecting, and so on. When a new node is processed and a full tree has
 * just been made, a half tree is created with the new node as the root,
 * and is put into the array. When we run out of nodes, we simply traverse
 * the array and stitch any remaining trees together.
 */

void udict_tree_convert_from_list(udict_t *ud)
{
    udict_node_t *tree[TREE_DEPTH_MAX] = { 0 };
    udict_node_t *curr, *nil = tree_null_priv(ud), *next;
    udict_node_t *complete = 0;
    udict_count_t fullcount = UDICT_COUNT_T_MAX, nodecount = ud->nodecount;
    udict_count_t botrowcount;
    int baselevel = 0, level = 0, i;

    while (fullcount >= nodecount && fullcount)
	fullcount >>= 1;

    botrowcount = nodecount - fullcount;

    for (curr = nil->next; curr != nil; curr = next) {
	next = curr->next;

	if (complete == 0 && botrowcount-- == 0) {
	    assert (baselevel == 0);
	    assert (level == 0);
	    baselevel = level = 1;
	    complete = tree[0];

	    if (complete != 0) {
		tree[0] = 0;
		complete->right = nil;
		while (tree[level] != 0) {
		    tree[level]->right = complete;
		    complete->parent = tree[level];
		    complete = tree[level];
		    tree[level++] = 0;
		}
	    }
	}

	if (complete == 0) {
	    curr->left = nil;
	    curr->right = nil;
	    complete = curr;

	    assert (level == baselevel);
	    while (tree[level] != 0) {
		tree[level]->right = complete;
		complete->parent = tree[level];
		complete = tree[level];
		tree[level++] = 0;
	    }
	} else {
	    curr->left = complete;
	    complete->parent = curr;
	    tree[level] = curr;
	    complete = 0;
	    level = baselevel;
	}
    }

    if (complete == 0)
	complete = nil;

    for (i = 0; i < TREE_DEPTH_MAX; i++) {
	if (tree[i] != 0) {
	    tree[i]->right = complete;
	    complete->parent = tree[i];
	    complete = tree[i];
	}
    }

    nil->right = nil;
    nil->left = complete;
    complete->parent = nil;
}

/*
 * Perform a ``left rotation'' adjustment on the tree.  The given parent node P
 * and its right child C are rearranged so that the P instead becomes the left
 * child of C.   The left subtree of C is inherited as the new right subtree
 * for P.  The ordering of the keys within the tree is thus preserved.
 */

void udict_tree_rotate_left(udict_node_t *child, udict_node_t *parent)
{
    udict_node_t *leftgrandchild, *grandpa;

    assert (parent->right == child);

    child = parent->right;
    parent->right = leftgrandchild = child->left;
    leftgrandchild->parent = parent;

    child->parent = grandpa = parent->parent;

    if (parent == grandpa->left) {
	grandpa->left = child;
    } else {
	assert (parent == grandpa->right);
	grandpa->right = child;
    }

    child->left = parent;
    parent->parent = child;
}


/*
 * This operation is the ``mirror'' image of rotate_left. It is
 * the same procedure, but with left and right interchanged.
 */

void udict_tree_rotate_right(udict_node_t *child, udict_node_t *parent)
{
    udict_node_t *rightgrandchild, *grandpa;

    assert (parent->left == child);

    parent->left = rightgrandchild = child->right;
    rightgrandchild->parent = parent;

    child->parent = grandpa = parent->parent;

    if (parent == grandpa->right) {
	grandpa->right = child;
    } else {
	assert (parent == grandpa->left);
	grandpa->left = child;
    }

    child->right = parent;
    parent->parent = child;
}


udict_operations_t udict_tree_ops = {
    udict_tree_init,
    udict_tree_insert,
    tree_delete,
    udict_tree_lookup,
    udict_tree_lower_bound,
    udict_tree_upper_bound,
    udict_tree_first,
    udict_tree_last,
    udict_tree_next,
    udict_tree_prev,
    udict_tree_convert_to_list,
    udict_tree_convert_from_list,
    udict_bst
};
