/*
 * Austin---Astonishing Universal Search Tree Interface Novelty
 * Copyright (C) 2000 Kaz Kylheku <kaz@ashi.footprints.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * This source code may be translated into executable form and incorporated
 * into proprietary software; there is no requirement for such software to
 * contain a copyright notice related to this source.
 *
 * $Id: udict.c,v 1.7 1999/12/09 07:32:48 kaz Exp $
 * $Name: austin_0_1 $
 */

#include <stdlib.h>
#include <assert.h>
#define UDICT_IMPLEMENTATION
#include "udict.h"
#include "macros.h"

extern udict_operations_t udict_list_ops;	/* see list.c */
extern udict_operations_t udict_tree_ops;	/* see tree.c */
extern udict_operations_t udict_redblack_ops;	/* see redblack.c */
extern udict_operations_t udict_splay_ops;	/* see splay.c */
extern udict_operations_t udict_avl_ops;	/* see avl.c */

static udict_operations_t *alg_table[] = {
    &udict_list_ops, &udict_tree_ops, &udict_redblack_ops, &udict_splay_ops,
    &udict_avl_ops
};

static udict_node_t *default_nodealloc(void *context)
{
    return malloc(sizeof *default_nodealloc(context));
}

static void default_nodefree(void *context, udict_node_t *node)
{
    free(node);
}

void udict_init(udict_t *ud, int algo, udict_count_t maxcount, udict_compare_t compare)
{
    ud->nodecount = 0;
    ud->maxcount = maxcount;
    ud->dupes_allowed = 0;
    ud->compare = compare;
    ud->nodealloc = default_nodealloc;
    ud->nodefree = default_nodefree;
    ud->context = 0;
    ud->optable = alg_table[algo];
    ud->optable->udict_init(ud);
}

udict_t *udict_create(int algo, udict_count_t maxcount, udict_compare_t compare)
{
    udict_t *newdict = malloc(sizeof *newdict);

    if (newdict)
	udict_init(newdict, algo, maxcount, compare);

    return newdict;
}

void udict_destroy(udict_t *ud)
{
    assert (udict_isempty(ud));
    free(ud);
}

void udict_convert_to(udict_t *ud, int algo)
{
    /*
     * Conversion to the existing type is a no-op
     */

    if (alg_table[algo] == ud->optable)
	return;

    /*
     * Conversion of any binary tree to a regular BST requires only a vtable
     * switch. Ditto if the target type is a tree that doesn't maintain
     * special balancing meta-information, such as any self-adjusting tree.
     */

    if (ud->optable->udict_kind == udict_bst && (algo == UDICT_BST || algo == UDICT_SPLAY)) {
	ud->optable = alg_table[algo];
	return;
    }


    /*
     * Otherwise, the source object is flattened to a list, and
     * the vtable is updated to the target type.
     */

    ud->optable->udict_convert_to_list(ud);
    ud->optable = alg_table[algo];

    /*
     * If the target type is a list, the job is done.
     */

    if (alg_table[algo]->udict_kind == udict_list)
	return;

    /*
     * If target type is not a list, we must tell it to rebuild itself
     * from a list.
     */

    ud->optable->udict_convert_from_list(ud);
}

udict_count_t udict_count(udict_t *ud)
{
    return ud->nodecount;
}

int udict_isempty(udict_t *ud)
{
    return ud->nodecount == 0;
}

int udict_isfull(udict_t *ud)
{
    return ud->nodecount == ud->maxcount;
}

int udict_alloc_insert(udict_t *ud, const void *key, void *data)
{
    udict_node_t *newnode = ud->nodealloc(ud->context);

    if (newnode != 0) {
	newnode->data = data;
	ud->optable->udict_insert(ud, newnode, key);
    }

    return newnode != 0;
}

void udict_delete_free(udict_t *ud, udict_node_t *node)
{
    ud->optable->udict_delete(ud, node);
    ud->nodefree(ud->context, node);
}

void udict_set_allocator(udict_t *ud, udict_nodealloc_t alloc, udict_nodefree_t dealloc, void *context)
{
    ud->nodealloc = alloc;
    ud->nodefree = dealloc;
    ud->context = context;
}

void udict_allow_dupes(udict_t *ud)
{
    assert (ud->nodecount == 0);
    ud->dupes_allowed = 1;
}

void udict_node_init(udict_node_t *node, void *data)
{
    node->data = data;
    node->algo_specific.udict_dummy = 0;
    node->left = 0;
    node->right = 0;
    node->parent = 0;
}

udict_node_t *udict_node_create(void *data)
{
    udict_node_t *newnode = malloc(sizeof *newnode);

    if (newnode != 0)
	newnode->data = data;
    
    return newnode;
}

void udict_node_destroy(udict_node_t *node)
{
    free(node);
}

void *udict_node_getdata(udict_node_t *node)
{
    return node->data;
}

void udict_node_setdata(udict_node_t *node, void *data)
{
    node->data = data;
}

const void *udict_node_getkey(udict_node_t *node)
{
    return node->key;
}

void udict_insert(udict_t *ud, udict_node_t *node, const void *key)
{
    ud->optable->udict_insert(ud, node, key);
}

void udict_delete(udict_t *ud, udict_node_t *node)
{
    ud->optable->udict_delete(ud, node);
}

udict_node_t *udict_lookup(udict_t *ud, const void *key)
{
    return ud->optable->udict_lookup(ud, key);
}

udict_node_t *udict_lower_bound(udict_t *ud, const void *key)
{
    return ud->optable->udict_lower_bound(ud, key);
}

udict_node_t *udict_upper_bound(udict_t *ud, const void *key)
{
    return ud->optable->udict_upper_bound(ud, key);
}

udict_node_t *udict_first(udict_t *ud)
{
    return ud->optable->udict_first(ud);
}

udict_node_t *udict_last(udict_t *ud)
{
    return ud->optable->udict_last(ud);
}

udict_node_t *udict_next(udict_t *ud, udict_node_t *node)
{
    return ud->optable->udict_next(ud, node);
}

udict_node_t *udict_prev(udict_t *ud, udict_node_t *node)
{
    return ud->optable->udict_prev(ud, node);
}
